;;======================================================================;;
;;			SERVO ESC DECODER				;;
;;======================================================================;;
;;									;;
;; Program:         SERVO_READ -- read servo signal	;;
;; Code:            Jindra Fucik				;;
;; Platform:        Microchip PIC16F628A, 4 Mhz				;;
;; Date:            04.03.2014						;;
;; First release:   04.03.2014						;;
;; LastDate:        24.08.2017						;;
;;									;;
;;======================================================================;;

; Minimal external components, uses internal oscilator at 4 MHz


; Servo working range 700usec - 2700usec
; with prescaller 1:16 will be 688/16=43 to 2704/16 = 169
; 169-43 = 126 hmm, 12F629 have 128B eeprom, then speed table can be in eeprom

; This program is distributed as is but WITHOUT ANY WARRANTY
; I hope you enjoy!!
;
; Revisions:
; 04.03.2014	Start of writting code
; 24.08.2017	changed to 12F629 + changed to ESC


; ----- Definitions

#define		__VERNUM	D'1'
#define		__VERDAY	0x04
#define		__VERMONTH	0x03
#define		__VERYEAR	0x14


	LIST	   p=12F629	; target processor

	;errorlevel -305
	errorlevel -302 ; warning, register in other bank

	#include p12F629.inc


	__CONFIG  _BODEN_ON & _CP_OFF & _WDT_OFF & _MCLRE_OFF & _PWRTE_OFF & _INTRC_OSC_NOCLKOUT 


;             +---------+
; Power   ---1|Vdd U Vss|8--- GND
; OUT A O ---2|GP5   GP0|7--- I limit A
; OUT B O ---3|GP4   GP1|6--- I Limit B
;       I ---4|GP3   INT|5--- I servo in
;             +---------+
#define		LIM_A	GPIO,0			; Input for move limiter A
#define		LIM_B	GPIO,1			; Input for move limiter B
#define		SRVIN	GPIO,2			; servo signal input pin
#define		XOUT_A	GPIO,4			; output A is on GPIO 4
#define		XOUT_B	GPIO,5			; output B is on GPIO 5

; --- Constant values

FXTAL		equ	D'4000000'		; internal oscilator

GP_TRIS		equ	b'00001111'		; GP2 servo input
OPTION_INI	equ	b'11000011'		; 01000011
                                ; 0-------, PORTB pull ups - enabled
                                ; -1------, Interrupt on rising edge of RB0/INT pin
                                ; --0-----, T0CS source Fosc/4
                                ; ---0----, T0SE edge lo>hi
                                ; ----0---, PSA prescale TMR0
                                ; -----011, PS prescaler 1:16

INTC_INI	equ	0x90			; GIE, INTE enable
PIE1_INI	equ	0x00			; no PIE
WPU_INI		equ	b'00001111'		; pill ups to inputs

T1CON_INI	equ b'00000001'		; Timer 1 on, 1:1 , internal 
TMR1H_INI	equ 0xFE			; FE = 254 = 512 cycles = 2khz base freq 16 steps = 122Hz repeat speed


; --- EEPROM Section

#define		EE_INI		0x00
; here is no strong EE definition
; pulse width-43 mean eeprom position - up to 169-43=129
; less than 43 mean 43, great than 169 mean 169 

; ----- Variables

; --- Internal RAM Section

        cblock  0x30				; all banks

INT_W		; interrupt context registers
INT_STAT	; interrupt status

EEDATA0	; data to write in EEPROM
        ;endc

        ;cblock  0x20				; bank 0

DATAVAL		; measured data
OLDDATAV	; OLD measured data
PWMCNT		; count of PWM steps
PWMVAL		; current PWM
;EEPWM		; PWM value readed from eeprom
NEWGPIO		; new value for GIPO for PWM

FLAGS		; various flags
TEMP		; temp values

        endc


; --- Flags

#define		NEW_DATA	FLAGS,0		; New packet received
#define		PWM_DIR		FLAGS,1		; Direction for PWM

#define		OUT_A		NEWGPIO,4	; output A is on GPIO 4
#define		OUT_B		NEWGPIO,5	; output B is on GPIO 5



; --------------- Program Section --------------------------------------


		org	0x000

PowerUp:
		clrf	STATUS			; Bank 0 default
		clrf	INTCON			; Disable all interrupts
		clrf	PCLATH			; tables on page 0
		goto	INIT

; ----------------------------------------------------------------------

		org	0x004

Interrupt:
		movwf	INT_W			; save context registers
		swapf	STATUS,w
		movwf	INT_STAT
		clrf	STATUS			; interrupt uses bank 0

Int_DCC:
		btfss	INTCON,INTF		; GPIO,2 interrupt?
		goto	EndInt

		btfsc	SRVIN
		goto	Int_High_Half
Int_Low_Half:
;		movlw	0x04
;		movwf	PORTB

		movf	TMR0,w
		movwf	DATAVAL
		btfss	INTCON,T0IF	; overflow = inconsistent signal
		bsf		NEW_DATA
		bcf	INTCON,T0IF		; clear overflow flag for counting

		bcf	INTCON,INTF		
		bsf	STATUS,RP0
		bsf	OPTION_REG,INTEDG	; next interrupt on rising edge RB0
		goto	EndInt

Int_High_Half:
;		movlw	0x08
;		movwf	PORTB

		clrf	TMR0		; prepare TMR0 for new mwasure
		bcf	INTCON,T0IF		; clear overflow flag for counting

		bcf	INTCON,INTF
		bsf	STATUS,RP0
		bcf	OPTION_REG,INTEDG	; next int. on falling edge RB0
EndInt:
		swapf	INT_STAT,w		; restore context registers
		movwf	STATUS
		swapf	INT_W,f
		swapf	INT_W,w
		retfie


; ----------------------------------------------------------------------


INIT:
		movlw	0x07
		movwf	CMCON			; set GP2:0 to digital I/O

		bsf	STATUS,RP0		; bank 1
		movlw	GP_TRIS
		movwf	TRISIO
		movlw	WPU_INI			; pull-ups
		movwf	WPU
		;clrf	IOC			; interrupt on change
		clrf	VRCON			; voltage reference off
		movlw	OPTION_INI		; Option register: no pull-up, falling GP2, no prescaler, wdt 1:1
		movwf	OPTION_REG
		movlw	PIE1_INI
		movwf	PIE1
		call	3FFh			;Get the cal value
		movwf	OSCCAL			;Calibrate
		bcf	STATUS,RP0			; bank 0
		clrf	PIR1
		movlw	T1CON_INI		; Timer 1 on, 1:1 , internal 
		movwf	T1CON
		movlw	TMR1H_INI		; Timer 1 512 steps 
		movwf	TMR1H

		movlw	0x20			; clear RAM
		movwf	FSR
ClearRAM:
		clrf	INDF
		incf	FSR,f
		movlw	0x60
		xorwf	FSR,w
		btfss	STATUS,Z
		goto	ClearRAM

		movlw	INTC_INI                ; Int enable, enable RA0 int, 
		movwf	INTCON					; enable RA0 external interrupt


; ----------------------------------------------------------------------

MainLoop:
		btfsc	NEW_DATA		; new data received?
		call	Decode			; yes, decode

		btfsc	PIR1,TMR1IF		; new pwm tick
		call	DoPWM			; yes, next PWM step

		goto	MainLoop

Decode:
		bcf		NEW_DATA

		movf	OLDDATAV,w		; compare old
		xorwf	DATAVAL,w		; and new value
		btfsc	STATUS,Z		; once they are same
		return					; do nothing

		movf 	DATAVAL,w
		movwf	OLDDATAV
		
		movlw	.43
		subwf	DATAVAL,w			; w=DATAVAL - 43
		btfss	STATUS,C		; DATAVAL is less than CVMAX
		movlw	0				; no - then is DATAVAL bigger or equal
 		movwf	TEMP			; store to TEMP

 		movlw	.127			; maximum value
 		btfsc	TEMP,7
 		movwf	TEMP

		movf	TEMP,w			; read PWM value from eeprom
		call	EE_Read
		movwf	PWMVAL

		bcf	PWM_DIR
		btfsc	PWMVAL,7
		bsf	PWM_DIR
		bcf	PWMVAL,7
		
		return


; ----------------------------------------------------------------------
DoPWM:
		movlw	TMR1H_INI		; Timer 1 512 steps 
		movwf	TMR1H
		bcf	PIR1,TMR1IF			; redy for new pwm tick
		movf	NEWGPIO,w
		movwf	GPIO
		
		incf PWMCNT,f				; increase global pwm counter
		movlw 0x0F					; number of PWM steps
		andwf PWMCNT,f			; reduce to proper amount of steps

		bcf OUT_A
		bcf OUT_B
		movf PWMVAL,w
		subwf PWMCNT,w			; w = COUNTPWM - PWM1CNT (if PWM1CNT > COUNTPWM then carry is set)
		btfsc STATUS,C
		return						; if not set, then do nothing, next is empty
		
		btfss	PWM_DIR				; is forward direction?
		bsf	OUT_A
		

		btfsc	PWM_DIR				; is forward direction?
		bsf	OUT_B

		btfss	LIM_A				; if limit is reach
		bcf	OUT_A					; do not allow move to this direction
		btfss	LIM_B				; if limit is reach
		bcf	OUT_B					; do not allow move to this direction
		
		return

; ----------------------------------------------------------------------

LoadCV:
		;movlw	E_MIN			; value low
		;call	EE_Read
		;movwf	CVMIN
		;movlw	E_MAX			; value high
		;call	EE_Read
		;movwf	CVMAX

		return


;---------------------------------------------------------------------------

EE_Read:
		bsf	STATUS,RP0		; w=ADR
		movwf	EEADR
		bsf	EECON1,RD
		movf	EEDATA,w
		bcf	STATUS,RP0
		return

SetParm:
		call	EE_Read			; w=ADR, EEDATA0=data. Write only changes
		xorwf	EEDATA0,w
		btfsc	STATUS,Z
		return
EE_Write:		
		movf	EEDATA0,w
		bsf	STATUS,RP0
		movwf	EEDATA
		bsf	EECON1,WREN
		bcf	INTCON,GIE
		movlw	0x55
		movwf	EECON2
		movlw	0xAA
		movwf	EECON2
		bsf	EECON1,WR
		bsf	INTCON,GIE
		bcf	EECON1,WREN
EEWrite0:
		btfsc	EECON1,WR
		goto	EEWrite0
		bcf	STATUS,RP0
		return



; ----- EEPROM default values


		org	0x2100

		dw	0x10			; 0
		dw	0x10			; 1
		dw	0x10			; 2
		dw	0x10			; 3
		dw	0x0F			; 4
		dw	0x0F			; 5
		dw	0x0F			; 6
		dw	0x0E			; 7
		dw	0x0E			; 8
		dw	0x0E			; 9
		dw	0x0D			; 10
		dw	0x0D			; 11
		dw	0x0D			; 12
		dw	0x0C			; 13
		dw	0x0C			; 14
		dw	0x0C			; 15
		dw	0x0B			; 16
		dw	0x0B			; 17
		dw	0x0B			; 18
		dw	0x0A			; 19
		dw	0x0A			; 20
		dw	0x0A			; 21
		dw	0x09			; 22
		dw	0x09			; 23
		dw	0x09			; 24
		dw	0x08			; 25
		dw	0x08			; 26
		dw	0x08			; 27
		dw	0x07			; 28
		dw	0x07			; 29
		dw	0x07			; 30
		dw	0x06			; 31
		dw	0x06			; 32
		dw	0x06			; 33
		dw	0x05			; 34
		dw	0x05			; 35
		dw	0x05			; 36
		dw	0x04			; 37
		dw	0x04			; 38
		dw	0x04			; 39
		dw	0x03			; 40
		dw	0x03			; 41
		dw	0x03			; 42
		dw	0x02			; 43
		dw	0x02			; 44
		dw	0x02			; 45
		dw	0x01			; 46
		dw	0x01			; 47
		dw	0x01			; 48
		dw	0x00			; 49
		dw	0x00			; 50
		dw	0x00			; 51 <--- usual centre
		dw	0x80			; 52
		dw	0x80			; 53
		dw	0x80			; 54
		dw	0x81			; 55
		dw	0x81			; 56
		dw	0x81			; 57
		dw	0x82			; 58
		dw	0x82			; 59
		dw	0x82			; 60
		dw	0x83			; 61
		dw	0x83			; 62
		dw	0x83			; 63
		dw	0x84			; 64
		dw	0x84			; 65
		dw	0x84			; 66
		dw	0x85			; 67
		dw	0x85			; 68
		dw	0x85			; 69
		dw	0x86			; 70
		dw	0x86			; 71
		dw	0x86			; 72
		dw	0x87			; 73
		dw	0x87			; 74
		dw	0x87			; 75
		dw	0x88			; 76
		dw	0x88			; 77
		dw	0x88			; 78
		dw	0x89			; 79
		dw	0x89			; 80
		dw	0x89			; 81
		dw	0x8A			; 82
		dw	0x8A			; 83
		dw	0x8A			; 84
		dw	0x8B			; 85
		dw	0x8B			; 86
		dw	0x8B			; 87
		dw	0x8C			; 88
		dw	0x8C			; 89
		dw	0x8C			; 90
		dw	0x8D			; 91
		dw	0x8D			; 92
		dw	0x8D			; 93
		dw	0x8E			; 94
		dw	0x8E			; 95
		dw	0x8E			; 96
		dw	0x8F			; 97
		dw	0x8F			; 98
		dw	0x8F			; 99
		dw	0x90			; 100
		dw	0x90			; 101
		dw	0x90			; 102
		dw	0x90			; 103
		dw	0x90			; 104
		dw	0x90			; 105
		dw	0x90			; 106
		dw	0x90			; 107
		dw	0x90			; 108
		dw	0x90			; 109
		dw	0x90			; 110
		dw	0x90			; 111
		dw	0x90			; 112
		dw	0x90			; 113
		dw	0x90			; 114
		dw	0x90			; 115
		dw	0x90			; 116
		dw	0x90			; 117
		dw	0x90			; 118
		dw	0x90			; 119
		dw	0x90			; 120
		dw	0x90			; 121
		dw	0x90			; 122
		dw	0x90			; 123
		dw	0x90			; 124
		dw	0x90			; 125
		dw	0x90			; 126
		dw	0x90			; 127

	end